package gc.globalGC;

import java.lang.ClassCastException;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Vector;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import wwc.naming.WWCNamingService;
import salsa.language.*;
import salsa.naming.UAL;
import salsa.naming.UAN;
import gc.*;
import gc.message.RemoveInverseRefMsg;
/**
 * <p>Title: local garbage collector for global garbage collection</p>
 * <p>Description: The local gabage finder uses the Pseudo-Root
 *      IDentifincation (PRID) algorithm. There are two kinds of
 *      pseudo roots. The first one is the active pseudo root. The
 *      second is the outside (the snapshot) pseudo root. </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: WWC</p>
 * @author WeiJen Wang
 * @version 1.0
 */

public class GLocalPRID implements LocalCollector{

  //the local naming service
  private WWCNamingService nameSvc;

  //the name of the global garbage tracing actor who occupy this local garbage collector
  //Querying with the name gets a lease time to release the lock.
  private Hashtable lock;

  //the actors in the snapshot
  private Hashtable actorArray;

  //the internal data structure for mark and sweep
  private LinkedList BFSMemVector;

  //the temporary storage for the system massage "RemoveInverseRefMsg"
  private Vector removeInvRefTarget;

  //the garbage array identified by the prid algorithm
  private Hashtable garbageArray;

  //the array stores the actor memories which are considerd as non-garbage
  //by remote reference.
  //******************************************************************************
  //*{non-local-reachable actors} = {live actors} - {local-pseudoroot-reachable actors}*
  //******************************************************************************
  private Hashtable nonlocalReachable;
  private Hashtable GNonlocalReachable;


  public class invTargetWrap {
    public String target;
    public String argument;
    int debt;
    public invTargetWrap(String target,String arg,int debt) {
      this.target=target;this.argument=arg;this.debt=debt;
    }
  }


  /****************************************
   * The following code is used for
   * the global cyclic garbage dective
   ****************************************/

  public synchronized boolean occupy(String detectiveName) {
      lock.put(detectiveName,new Long(System.currentTimeMillis()));
      return false;
  }

  public synchronized void free(String detectiveName) {
    lock.remove(detectiveName);
  }

  public synchronized Hashtable getNonlocalReachableSet(String detectiveName) {
    return GNonlocalReachable;
  }


  private synchronized void preserveNonlocalSet() {
    if (lock.size()>0) {return;}
    GNonlocalReachable=nonlocalReachable;
    nonlocalReachable=new Hashtable();
  }


  ActorMemory forTesting;
  private synchronized void testing() {
    if (forTesting==null) {return;}
    GCAgent gcAgent=new GCAgent(null,null);
    WeakReference gcWeakRef=new WeakReference(gcAgent);
    GBackTracer gbt=new GBackTracer(forTesting);
    Object[] args={gbt};
    gcAgent.send(new Message(gcWeakRef,gcWeakRef,"construct",args,false));
  }

  public int getThroughPut() {return 0;}

  /***************************************
   * The following code is impemented for the
   * local garbage identifier.
   ***************************************/
  public GLocalPRID() {
    try {
      nameSvc = (WWCNamingService) ServiceFactory.getNaming();
    } catch (ClassCastException e) {nameSvc=null;}
    actorArray=new Hashtable();
    BFSMemVector=new LinkedList();
    removeInvRefTarget=new Vector();
    garbageArray=new Hashtable();
    nonlocalReachable=new Hashtable();
    GNonlocalReachable=new Hashtable();
    lock=new Hashtable();
  }

  public void collect() {
    snapshot();
    localPseudoRootIdentification();
    BFSMark();
    nonLocalPseudoRootIdentification();
    nonLocalBFSMark();
    getGarbageActorAndAcqInvRef();
    collectGarbage();
    preserveNonlocalSet();
    GCFinalize();
    testing();
  }

  protected void snapshot() {
    snapshotUANReg(nameSvc.getUANTable());
    snapshotUALReg(nameSvc.getUALTable());
    removeInvRefTarget.clear();
    garbageArray.clear();
  }

  protected void localPseudoRootIdentification() {
      BFSMemVector.clear();
      for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
        UniversalActor.State actor=(UniversalActor.State)e.nextElement();
        ActorMemory currentMemory=actor.getActorMemory();

        //if the actor is unblocked, color it
        if (currentMemory.isActive() || currentMemory.isLocked()) {
          currentMemory.colorPseudoRoot();
          BFSMemVector.addLast(currentMemory);
        }

        //if it is wating for an ACK, color it.
        else if (currentMemory.GCGetForwardList().isPseudoRoot()) {
          currentMemory.colorPseudoRoot();
          BFSMemVector.addLast(currentMemory);
        }
      }
  }

  protected void nonLocalPseudoRootIdentification() {
      BFSMemVector.clear();
      for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
        UniversalActor.State actor=(UniversalActor.State)e.nextElement();
        ActorMemory currentMemory=actor.getActorMemory();

        if (currentMemory.isMarked() || currentMemory.isPseudoRoot()) {continue;}

        // One step back tracking:
        //   If an actor has an inverse acquaintance outside
        //   the snapshot group, color it.
        else {
          Hashtable ilist = currentMemory.GCGetinverseList().getRealInverseList();
          if (ilist==null) {continue;}
          for (Enumeration invRefStrEnu = ilist.keys();
               invRefStrEnu.hasMoreElements(); ) {
            String searchKey = (String) invRefStrEnu.nextElement();
            UniversalActor.State inverseAcquaintance=(UniversalActor.State)actorArray.get(searchKey);
            if (inverseAcquaintance == null) {
              currentMemory.colorNonLocalPseudoRoot();
              BFSMemVector.addLast(currentMemory);
              nonlocalReachable.put(currentMemory.GCGetForwardList().getSelfRef(),currentMemory);
              break;
            }
          }
        }
      }
  }


  protected void BFSMark() {
    ActorMemory currentMem;
    for (;;) {

      try {currentMem = (ActorMemory) BFSMemVector.getFirst();}
      catch (NoSuchElementException e) {break;}


      Hashtable flist=currentMem.GCGetForwardList().getRealForwardList();
      for (Enumeration e = flist.keys();  e.hasMoreElements(); ) {
        String nextActorStr=(String)e.nextElement();
        ForwardListEntry fEntry=(ForwardListEntry)flist.get(nextActorStr);

        if (fEntry.getReferenceCounter()<=0) {continue;}
        Object nextActorObj=actorArray.get(nextActorStr);
        if (nextActorObj!=null) {
          UniversalActor.State nextActor = (UniversalActor.State) nextActorObj;
          ActorMemory nextMem = nextActor.getActorMemory();
          if (nextMem.isLocalMarked() || nextMem.isPseudoRoot()) {
            continue;
          } else {
            nextMem.colorLocalMark();
            BFSMemVector.addLast(nextMem);
          }
        }
      }
      try {BFSMemVector.removeFirst();}
      catch (ArrayIndexOutOfBoundsException e) {break;}
    }
    BFSMemVector.clear();
  }

  protected void nonLocalBFSMark() {
    ActorMemory currentMem;
    for (;;) {
      try {currentMem = (ActorMemory) BFSMemVector.getFirst();}
      catch (NoSuchElementException e) {break;}

      Hashtable flist=currentMem.GCGetForwardList().getRealForwardList();
      for (Enumeration e = flist.keys();  e.hasMoreElements(); ) {
        String nextActorStr=(String)e.nextElement();
        ForwardListEntry fEntry=(ForwardListEntry)flist.get(nextActorStr);

        if (fEntry.getReferenceCounter()<=0) {continue;}
        Object nextActorObj=actorArray.get(nextActorStr);
        if (nextActorObj!=null) {
          UniversalActor.State nextActor = (UniversalActor.State) nextActorObj;
          ActorMemory nextMem = nextActor.getActorMemory();
          if (nextMem.isLocalMarked() || nextMem.isPseudoRoot() ||
              nextMem.isNonLocalMarked() || nextMem.isNonLocalPseudoRoot()) {
            continue;
          } else {
            nextMem.colorNonLocalMark();
            BFSMemVector.addLast(nextMem);
            nonlocalReachable.put(nextMem.GCGetForwardList().getSelfRef(),nextMem);
forTesting=nextMem;
          }
        }
      }
      try {BFSMemVector.removeFirst();}
      catch (ArrayIndexOutOfBoundsException e) {break;}
    }
    BFSMemVector.clear();
  }


  protected void getGarbageActorAndAcqInvRef() {
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State garbageActor=(UniversalActor.State) e.nextElement();
      ActorMemory currentMem=garbageActor.getActorMemory();
      if (!currentMem.isLocalMarked() && !currentMem.isPseudoRoot() &&
      !currentMem.isNonLocalMarked() && !currentMem.isNonLocalPseudoRoot()) {
        garbageArray.put(currentMem.GCGetForwardList().getSelfRef(),garbageActor);
        removeAcquaintanceInvRef(currentMem.GCGetForwardList(),true);
      }
      removeAcquaintanceInvRef(currentMem.GCGetForwardList(),false);
    }
  }


  //try to remove useless forward references and put them in 'removeInvRefTarget'
  private void removeAcquaintanceInvRef(ForwardList fList, boolean killed) {
    //find all unused forward references
    for (Enumeration e = fList.getRealForwardList().keys();  e.hasMoreElements(); ) {
      String key=(String) e.nextElement();
      Hashtable flistTable=fList.getRealForwardList();
      ForwardListEntry fEntry=(ForwardListEntry) flistTable.get(key);

      //if the forward reference is useless, remove it, and store their information
      // the forward references are useless if
      //  1. the source actor is live but some of its forward references are dead
      //  2. the source actor is garbage.
      if (fEntry.isDead() || killed) {
         Object res=flistTable.remove(key);
         if (res!=null) {
           //put the inverse reference information in 'removeInvRefTarget'
           invTargetWrap invData=new invTargetWrap(key,fList.getSelfRef(),fEntry.getDebt());
           removeInvRefTarget.addElement(invData);
         }
      }
    }
  }

  protected void collectGarbage() {
    for (Enumeration e = garbageArray.elements();  e.hasMoreElements();) {
      UniversalActor.State gActor=((UniversalActor.State)e.nextElement());
      nameSvc.remove(gActor.getUAN(),gActor.getUAL());
      gActor.GCdie();
    }
  }

  protected void GCFinalize() {
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State actor = (UniversalActor.State)e.nextElement();
      try {actor.getActorMemory().endGC();}
      catch (Exception exc) {
        System.err.println("GC debug info:"+e);
        continue;
      }
    }
    notifyInvAcq();
    this.removeInvRefTarget.clear();
    this.garbageArray.clear();
this.showActorState();
    this.actorArray.clear();
    this.BFSMemVector.clear();
    System.gc();
  }

  /**
   * firstly, remove all unnecessary msgs whose targets are garbage.
   * Then start sending messages to unregister inverse references
   * This function is used in GCFinalize()
   */
  private void notifyInvAcq() {
    //remove all unnecessary system messages which will be sent to a garbage actor
    for (int i=0;i<removeInvRefTarget.size();i++) {
      invTargetWrap data=(invTargetWrap)removeInvRefTarget.get(i);
      if (garbageArray.get(data.target)==null) {
        //send system messages to the inverse acquaintances
        //to remove the inverse references
        WeakReference invActor;
        if (data.target.charAt(0)=='u') {invActor = new WeakReference(new UAN(data.target), null);}
        else {invActor = new WeakReference(null, new UAL(data.target));}
        invActor.send(new RemoveInverseRefMsg(invActor,data.argument, data.debt));
      }
    }
  }



  //pass the uan registry hashtable that stores actors
  // to find them and put them in the 'actorArray'
  private void snapshotUANReg(Hashtable reg) {
      for (Enumeration e = reg.elements();  e.hasMoreElements(); ) {
        Object actor = e.nextElement();
        if (actor instanceof UniversalActor.State) {
          if (actor instanceof salsa.resources.EnvironmentalServiceState) {continue;}
          if (actor instanceof salsa.resources.ActorServiceState) {continue;}
          actorArray.put(((UniversalActor.State)actor).getUAN().toString(),actor);
          ( (UniversalActor.State) actor).getActorMemory().startGC();
        }
      }
  }

  private void snapshotUALReg(Hashtable reg) {
      for (Enumeration e = reg.elements();  e.hasMoreElements(); ) {
        Object actor = e.nextElement();
        if (actor instanceof UniversalActor.State) {
          if (actor instanceof salsa.resources.EnvironmentalServiceState) {continue;}
          if (actor instanceof salsa.resources.ActorServiceState) {continue;}
          actorArray.put(((UniversalActor.State)actor).getUAL().toString(),actor);
          ( (UniversalActor.State) actor).getActorMemory().startGC();
        }
      }
  }


  //debugging tool:
  public void showActorState() {
    for (Enumeration e = actorArray.elements();  e.hasMoreElements(); ) {
      UniversalActor.State actor=(UniversalActor.State)e.nextElement();
      System.out.println(actor.getActorMemory().toString());
    }
  }
}
